/* CubicGraph.java
   -------------
   2002 (c) Dries Van Dyck
   Combinatorial Algorithms Research Group
   University of Ghent
   Krijgslaan 281--S9
   B-9000 GENT Belgium
*/

import java.io.IOException;
import java.io.BufferedReader;
import java.io.FileReader;
import java.util.StringTokenizer;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Hashtable;
import java.util.Arrays;
import java.util.NoSuchElementException;

/** 
 * This class represents cubic graphs implementing the {@link Yutsis}
 * interface.
 * @see Yutsis
 * @author Dries.VanDyck@rug.ac.be
 **/

public class CubicGraph implements Cloneable, Yutsis{

    /** 
     * Int representing a format in which the graphs are stored in 
     * a file: the first line contains the (cubic) order (n) of the graph,
     * 2n lines with three labels representing the couplings in the 
     * nodes. The first n lines correspond to the left tree in the braket,
     * the next n lines those of the right tree in the braket.
     */
    public static final int YTS = 1;

    /** 
     * Int representing a format in which the graphs are stored in a file:
     * this represents the well known braket notation e.g. 
     * <(j1,j2)j12|(j2,j1)j12>.
     */
    public static final int BRAKET = 2;

    /** 
     * Int representing a format in which the graphs are stored in a file:
     * <pre>
     * &lt;nrOfnodes&gt &lt;nrOfEdges&gt 
     * &lt;edge&gt &lt;edge&gt ... &lt;edge&gt
     * </pre>
     * With &lt;edge&gt = node1 node2, the edges may be distributed over 
     * several lines, but linebreaks are not allowed between two nodes
     * of the same edge. No sytax checks are performed. 
     */
    public static final int EDGELIST = 4;
    
    /** 
     * Adjacency table.
     */
    protected int [][] adjTab = null;
    
    /** 
     * Current cubic order, (2n nodes, 3n edges).
     */
    protected int n;

    /** 
     * Creates a new CubicGraph from a String in BRAKET format.
     */
    public CubicGraph(String braket){
	parseBraket(braket);
    }

    /**
     * Creates a new Cubic Graph which is read from the BufferedReader
     * in the given format:
     * @param input the source from where the CubicGraph is read
     * @param format the format of the CubicGraph in input, which is 
     *               {@link #BRAKET} or {@link #YTS} 
     */
    public CubicGraph(BufferedReader input, int format) throws IOException{
	StringTokenizer st;
	Hashtable edgesht;
	int [] freqtab;
	switch (format){
	case YTS:
	    st = new StringTokenizer(input.readLine());
	    if (st.countTokens() > 1)
		st.nextToken();
	    n = Integer.parseInt(st.nextToken());
	    edgesht = new Hashtable(4*n);
	    freqtab = new int[2*n];
	    adjTab = new int[2*n][3];
	    for (int node = 0; node < 2*n; node++){
		st = new StringTokenizer(input.readLine());
		for (int i = 0; i < 3; i++){
		    String label = st.nextToken();
		    Integer value = (Integer) edgesht.remove(label);
		    if (value == null){
			edgesht.put(label,new Integer(node));
		    }
		    else{
			int firstnode = value.intValue();
			adjTab[firstnode][freqtab[firstnode]++] = node;
			adjTab[node][freqtab[node]++] = firstnode;
		    }
		}
	    }
	    break;
	case BRAKET:
	    String line = input.readLine();
	    parseBraket(line);
	    break;
	case EDGELIST:
	    // Format: edgelist
	    st = new StringTokenizer(input.readLine());
	    n = Integer.parseInt(st.nextToken())/2;
	    adjTab = new int[2*n][3];
	    freqtab = new int[2*n];
	    st = new StringTokenizer(input.readLine());
	    for (int i=0; i < 3*n; ++i){
		if (!st.hasMoreTokens())
		    st = new StringTokenizer(input.readLine());
		int node1 = Integer.parseInt(st.nextToken());
		int node2 = Integer.parseInt(st.nextToken());
		adjTab[node1][freqtab[node1]++] = node2;
		adjTab[node2][freqtab[node2]++] = node1;
	    }   
	    break;
	default:
	    throw new IllegalArgumentException("Unknown format: " + format);
	}
    }
    
    private void parseBraket(String braket){
	StringTokenizer st = new StringTokenizer(braket,"<|>");
	String bra,ket;
	if (st.hasMoreTokens())
	    bra = st.nextToken().trim();
	else 
	    throw new IllegalArgumentException("BRA is empty (<BRA|KET>)");
	if (st.hasMoreTokens())
	    ket = st.nextToken().trim();
	else 
	    throw new IllegalArgumentException("KET is empty (<BRA|KET>)");
	if (bra.charAt(0) != '(')
	    throw new IllegalArgumentException("<\n"+bra+"| : expected '('");
	if (ket.charAt(0) != '(')
	    throw new IllegalArgumentException("<\n"+ket+"| : expected '('");
	int order = 0;
	for (int i = 0; i < bra.length(); i++)
	    if (bra.charAt(i) == '(')
		order++;
	n = 0;
	Hashtable edgesht = new Hashtable(4*order);
	int [] freqtab = new int[2*order];
	adjTab = new int[2*order][3];
	int pos = parseBraket(bra,0,true,edgesht,freqtab);
	String rest = bra.substring(pos).trim(); // should be the root label
	pos = parseBraket(ket,0,false,edgesht,freqtab);
	String root = ket.substring(pos).trim(); // should be the root label
	if (!root.equals(rest))
	    throw new IllegalArgumentException
		("Root labels from both trees differ:" + rest + "!=" + root);
	n = n/2;
    }

    private int parseBraket(String braket, int pos, boolean bra,
			    Hashtable edgesht, int [] freqtab){
	pos++;
	pos = skipSpaces(braket,pos);
	if (braket.charAt(pos) == '(')
	    pos = parseBraket(braket,pos,bra,edgesht,freqtab);
	int nextpos = braket.indexOf(',',pos);
	if (nextpos == -1)
	    throw new IllegalArgumentException
		((bra ? "<" : "|") + braket.substring(0,pos)+'\n'
		 + braket.substring(pos) + (bra ? "|" : ">")
		 + ": expected label followed by ','");
	String [] label = new String[3];
	label[0] = braket.substring(pos,nextpos).trim();
	    
	pos = ++nextpos;
	pos = skipSpaces(braket,pos);
	if (braket.charAt(pos) == '(')
	    pos = parseBraket(braket,pos,bra,edgesht,freqtab);
	nextpos = braket.indexOf(')',pos);
	if (nextpos == -1)
	    throw new IllegalArgumentException
		((bra ? "<" : "|") + braket.substring(0,pos)+'\n'
		 + braket.substring(pos) + (bra ? "|" : ">")
		 + ": expected label followed by ')'");
	label[1] = braket.substring(pos,nextpos).trim();
	pos = ++nextpos; // just behind ')'
	pos = skipSpaces(braket,pos);
	while (nextpos < braket.length() && braket.charAt(nextpos) != ')' 
	       && braket.charAt(nextpos) != ',')
	    nextpos++;
	label[2] = braket.substring(pos,nextpos).trim();
	for (int i =0 ; i < 3; i++){
	    Integer value = (Integer) edgesht.remove(label[i]);
	    if (value == null){
		edgesht.put(label[i], new Integer(n));
	    }
	    else {
		int firstnode = value.intValue();
		adjTab[firstnode][freqtab[firstnode]++] = n;
		adjTab[n][freqtab[n]++] = firstnode;
	    }
	}
	n++;
	return pos;
    }

    private int skipSpaces(String s, int pos){
	while (s.charAt(pos) == ' ')
	    pos++;
	return pos;
    }

    /**
     * The Copy Constructor: a copy of c is created.
     * @param c the CubicGraph to be duplicated
     */
    public CubicGraph(CubicGraph c){
	n = c.n;
	adjTab = new int[c.adjTab.length][3];
	for (int i = 0; i < adjTab.length; i++)
	    for (int j = 0; j < 3; j++)
		adjTab[i][j] = c.adjTab[i][j];
    }

    /**
     * Returns the (original order) of the graph, as specified by the
     * {@link Graph} interface.
     * @return the original order of the graph
     * @see Graph
     */
    public int order(){
	return adjTab.length;
    }
    
    /**
     * Returns the number of nodes of the graph, as specified by the
     * {@link Graph} interface.
     * @return the number of nodes
     * @see Graph
     */
    public int nrOfNodes(){
	return 2*n;
    }
    
    /**
     * Returns the number of edges of the graph, as specified by the
     * {@link Graph} interface. Note that this is always equal to
     * 3*nrOfNodes()/2. 
     * @return the number of edges
     * @see Graph 
     */
    public int nrOfEdges(){
	return 3*n;
    }

    /**
     * Returns the neighbors of <code>node</code> in the graph as an
     * array of ints, as specified by the {@link Graph} interface.  
     * @param node the node of which the neighbors will be returned
     * @return the neighbors of <code>node</code>
     */
    public int[] neighbors(int node){
	int [] nghb = new int[3];
	for (int i = 0; i < 3; ++i)
	    nghb[i] = adjTab[node][i];
	return nghb;
    }

    /** 
     * Class implementing the {@link NodeIterator} interface.
     */
    class CNodeIterator implements NodeIterator{
	/** 
	 * If <code>node == i</code> we see the nodeiterator as
	 * between the nodes <code>i</code> and the next node that is
	 * not removed.
	 */
	int node = -1;
	
	CNodeIterator(){}
	
	CNodeIterator(int node){
	    this.node = node;
	}
	
	public boolean hasNext(){
	    for(int i=this.node+1;i < order(); i++)
		if (!isRemoved(i))
		    return true;
	    return false;
	}

	public int next(){
	    for(node++;node < order(); node++)
		if (!isRemoved(node))
		    return node;
	    throw new NoSuchElementException();
	}

	public boolean hasPrevious(){
	    for(int i=node;i >= 0; i--)
		if (!isRemoved(i))
		    return true;
	    return false;
	}

	public int previous(){
	    for(; node >= 0; node--)
		if (!isRemoved(node))
		    return node--;
	    throw new NoSuchElementException();
	}
    }

    /** 
     * Returns an {@link NodeIterator} over the nodeset starting at the node 
     * with lowest label as specified by the
     * {@link Graph} interface.
     * @return a {@link NodeIterator} pointing to the first node
     * @see NodeIterator
     * @see Graph
     */
    public NodeIterator first(){
	return new CNodeIterator(-1);
    }

    /** 
     * Returns an {@link NodeIterator} over the nodeset starting at the node 
     * with highest label in some ordening as specified by the
     * {@link Graph} interface.
     * @return a {@link NodeIterator} pointing to the last node
     * @see NodeIterator
     * @see Graph
     */
    public NodeIterator last(){
	return new CNodeIterator(order()-1);
    }

    /** 
     * Returns an {@link NodeIterator} which will return <code>node</code> 
     * by a {@link NodeIterator#next()} call as specified by the
     * {@link Graph} interface.
     * @param node the node to be returned by the first 
     *             {@link NodeIterator#next()} of the returned 
     * 	           {@link NodeIterator}
     * @return a {@link NodeIterator} returning <code>node</code> by the
     *         first {@link NodeIterator#next()} call.
     * @see NodeIterator
     * @see Graph
     */
    public NodeIterator nodeIterator(int node){
	if (0 > node || node > order()-1 || isRemoved(node))
	    return null;
	for (int i = node-1; i >= 0 ; i--)
	    if (!isRemoved(i))
		return new CNodeIterator(i);
	return new CNodeIterator(node-1); // same effect as -1
    }

    /** 
     * True if <code>i</code> and <code>j</code> are connected and
     * false otherwise as specified by the {@link Graph}
     * interface.
     * @param i a node 
     * @param j a node
     * @return true if <code>i</code> and <code>j</code> are
     *         connected, false otherwise 
     * @see Graph
     */
    public boolean connected(int i, int j){
	if (isRemoved(i) || isRemoved(j))
	    return false;
	for(int k = 0; k < 3; k++)
	    if (adjTab[i][k] == j)
		return true;
	return false;
    }

    /** 
     * Returns the first bubble encoutered, or {-1,-1} if no bubbles
     * are present, as specified by the {@link Yutsis} interface.
     * @return the nodes of the bubble as an array or {-1,-1} if no bubbles
     *         are present
     * @see Yutsis
     */
    public int [] bubble(){
	int [] bubble = {-1,-1};
	for (int i = 0; i < order()-1; i++)
	    if (!isRemoved(i)){
		int [] nghb = neighbors(i);
		if (nghb[0]==nghb[1] || nghb[0] == nghb[2] ){
		    bubble[0]=i;
		    bubble[1]=nghb[0];
		    break;
		}
		else if (nghb[1] == nghb[2]){
		    bubble[0] = i;
		    bubble[1] = nghb[1];
		    break;
		}
	    }
	return bubble;
    }

    /** 
     * Interchange on the edge <code>(nodes[0], nodes[1])</code>
     * interchanging the edges <code>(nodes[0], icnodes[0])</code>
     * and <code>(nodes[1],icnodes[1])</code>, as specified by the
     * {@link Yutsis} interface.
     * @param nodes array containing the endpoints of the edge on
     *              which the interchange is performed
     * @param icnodes array containing the neighbors of the corresponding
     *                nodes of nodes to be interchanged 
     * @see Yutsis
     */
    public void interchange(int [] nodes, int [] icnodes){
	for (int i = 0; i < 2; i++)
	    for (int j = 0; j < 3; j++){
		if (adjTab[nodes[i]][j] == icnodes[i])
		    adjTab[nodes[i]][j] = icnodes[1-i]; // 1-i == (i+1) %2
		if (adjTab[icnodes[i]][j] == nodes[i])
		    adjTab[icnodes[i]][j] = nodes[1-i];
	    }
    }
    
    /** 
     * Interchange on the edge <code>(node1, node2)</code>
     * interchanging the edges <code>(node1, icnode1)</code> and
     * <code>(node2, icnode2)</code>, as specified by the {@link
     * Yutsis} interface.
     * @param node1 first node of the edge on which the interchange 
     *              is performed
     * @param node2 second node of the edge on which the interchange 
     *              is performed
     * @param icnode1 neighbor of node1 to be interchanged with icnode2
     * @param icnode2 neighbor of node2 to be interchanged with icnode1 
     * @see Yutsis
     */
    public void interchange(int node1, int node2, int icnode1, int icnode2){
	int [] nodes = {node1, node2};
	int [] icnodes = {icnode1, icnode2};
	interchange(nodes, icnodes);
    }

    /** 
     * Removes the bubble, as specified by the {@link Yutsis} interface.
     * @param bubble the bubble to be removed
     * @see Yutsis
     */
    public void removeBubble(int [] bubble){
	int [] nghb = {-1,-1};
	for (int i = 0; i < 3; i++){
	    if (adjTab[bubble[0]][i] != bubble[1])
		nghb[0] = adjTab[bubble[0]][i];
	    if (adjTab[bubble[1]][i] != bubble[0])
		nghb[1] = adjTab[bubble[1]][i];
	}
	for (int i = 0; i < 3; i++){
	    if (adjTab[nghb[0]][i] == bubble[0])
		adjTab[nghb[0]][i] = nghb[1];
	    if (adjTab[nghb[1]][i] == bubble[1])
		adjTab[nghb[1]][i] = nghb[0];
	}
	removeNodes(bubble[0],bubble[1]);
    }

    /** 
     * Removes the triangle <code>triangle[0], triangle[1],
     * triangle[2]</code>, as specified by the {@link Yutsis}
     * interface.
     * @param triangle array containing the nodes of the triangle to 
     *                 be removed 
     * @see Yutsis
     */
    public void removeTriangle(int [] triangle){
	int [] nghb = new int[3]; 
	for (int i = 0; i < 3; i++)
	    for (int j = 0; j < 3; j++){
		if (adjTab[triangle[i]][j] != triangle[(i+1) % 3] 
		    && adjTab[triangle[i]][j] != triangle[(i+2) % 3])
		    nghb[i] = adjTab[triangle[i]][j];
	    }
	for (int i = 0; i < 3; i++){
	    adjTab[triangle[0]][i] = nghb[i];
	    for (int j = 0; j < 3; j++)
		if (adjTab[nghb[i]][j] == triangle[i])
		    adjTab[nghb[i]][j] = triangle[0];
	}
	removeNodes(triangle[1],triangle[2]);
    }

    /** 
     * Removes the triangle <code>node1, node2, node3</code>, as
     * specified by the {@link Yutsis} interface.
     * @param node1 first node of the triangle to be removed
     * @param node2 second node of the triangle to be removed
     * @param node3 third node of the triangle to be removed 
     * @see Yutsis
     */
    public void removeTriangle(int node1, int node2, int node3){
	int [] triangle = {node1, node2, node3};
	removeTriangle(triangle);
    }

    /** 
     * Returns the first triangle encoutered, or {-1,-1,-1} if no
     * triangles are present, as specified by the {@link Yutsis}
     * interface.
     * @return the nodes of the triangle as an array or {-1,-1,-1} if
     *         no triangle are present.
     * @see Yutsis
     */
    public int [] triangle(){
	for (int i = 0; i < order()-2; i++)
	    if (!isRemoved(i)){
		int [] nghb = neighbors(i);
		for (int j = 0; j < 2; j++){
		    int [] nghbj = neighbors(nghb[j]);
		    for (int k = 0; k < 3; k++)
			if (nghbj[k] == nghb[j+1] 
			    || nghbj[k] == nghb[(j+2)%3])
			    if (nghb[j] < nghbj[k])
				// return it like CycleGenerator
				return new int [] {i, nghb[j], nghbj[k]};
			    else 
				return new int [] {i, nghbj[k], nghb[j]};
		}
	    }
	return new int [] {-1,-1,-1};
    }

    /** 
     * Implementation of the {@link Yutsis} interface.
     * @return all triangles as an array of int [3] objects.
     * @see Yutsis
     */
    public int [] [] triangles(){
	ArrayList triangles = new ArrayList();
	for (int i = 0; i < order()-2; i++)
	    if (!isRemoved(i)){
		int [] nghb = neighbors(i);
		for (int j = 0; j < 3; j++)
		    if (i < nghb[j]){
			int [] nghbj = neighbors(nghb[j]);
			for (int k = 0; k < 3; k++)
			    if (i < nghbj[k] && (nghbj[k] == nghb[(j+1)%3] 
				|| nghbj[k] == nghb[(j+2)%3]))
				if (nghb[j] < nghbj[k])
				    triangles.add(new int [] 
					{i, nghb[j], nghbj[k]});
		    }
	    }
	int [] [] toreturn = new int [triangles.size()][];
	triangles.toArray(toreturn);
	return toreturn;
    }

    /** 
     * True if the graph represents a triangular delta, false
     * otherwise, as specified by the {@link Yutsis} interface.
     * @return true if this graph represents a triangular delta 
     * @see Yutsis*/
    public boolean triangularDelta(){
	return n == 1;
    }

    /** 
     * True if the graph represents a triangular delta between node
     * <code>td[0]</code> and <code>td[1]</code>, false otherwise, as
     * specified by the {@link Yutsis} interface.
     * @param td the nodes of the triangular delta as an array
     * @return true if this Graph is a triangular delta between
     *         the specified nodes 
     * @see Yutsis
     */
    public boolean triangularDelta(int [] td){
	if (!isRemoved(td[0]) && !isRemoved(td[1]) && n == 1)
	    return true;
	return false;
    }

    /** 
     * True if the graph represents a triangular delta between node
     * <code>node1</code> and <code>node2</code>, false otherwise, as
     * specified by the {@link Yutsis} interface.
     * @param node1 first node of the triangular delta to be formatted
     * @param node2 second node of the triangular delta to be formatted
     * @return true if this Graph is a triangular delta between
     *         the specified nodes 
     * @see Yutsis
     */
    public boolean triangularDelta(int node1, int node2){
	int [] td = {node1, node2};
	return triangularDelta(td);
    }


    /** 
     * Returns true if the node <code>node</code> is removed, false
     * otherwise, as specified by the {@link Yutsis} interface.
     * @param node the node to be checked
     * @return true if the node is removed, false otherwise
     * @see Yutsis
     */ 
    boolean isRemoved(int node){ return adjTab[node][0] == -1;}
    
    /** 
     * Removes the nodes <code>node1</code> and <code>node2</code>.
     * @param node1 first node to be removed
     * @param node2 second node to be removed
     */
    private void removeNodes(int node1, int node2){
	adjTab[node1][0] = adjTab[node2][0] = -1;
	n--;
    }
    
    /**
     * Returns the girth, IF AND ONLY IF the graph is in maximal 
     * representation and IF the adjacency list are sorted ascendig,
     * otherwise an meaningless result is returned.
     * @return the girth of the graph
     */
    public int girthM(){
	int girth = 1;
	int prevnode = 0, node = 1, newnode=-1;
	for (int [] nghb = neighbors(1);
	     node != 0 && girth < adjTab.length; ++girth){
	    newnode = nghb[nghb[0] != prevnode ? 0 : 1];
	    prevnode = node; node = newnode;
	    nghb = neighbors(node);
	}
	if (girth == adjTab.length || prevnode != 2) // graph not maximal
	    girth = -1;
	return girth;
    }
    
    /**
     * Checks of the nodes in the nodeset form a vertex induced tree.
     * Returns also false when the nodeset is not connected.
     * @param nodeset the nodes to be checked if they induce a tree
     * @return true, if the specified nodeset induces a tree, false otherwise
     */
    public boolean isTree(int [] nodeset){
	int n = adjTab.length;
	boolean [] intree = new boolean[n]; // nodes in tree
	boolean [] inset = new boolean[n]; // nodes in the set
	for (int i = 0; i < nodeset.length; ++i)
	    inset[nodeset[i]] = true;
	int nrofnodesintree = 1;
	intree[nodeset[0]] = true;
	LinkedList ll = new LinkedList();
	ll.add(new Integer(nodeset[0]));
	while(!ll.isEmpty() && nrofnodesintree < n/2){
	    int currentnode = ((Integer) ll.removeFirst()).intValue();
	    int [] nghb = neighbors(currentnode);
	    for (int i = 0; i < 3; ++i)
		if (!intree[nghb[i]] && inset[nghb[i]]){
		    int [] newnghb = neighbors(nghb[i]);
		    for (int j = 0; j < 3; ++j)
			if (newnghb[j] != currentnode && intree[newnghb[j]])
			    return false; // Cycle!
		    ll.add(new Integer(nghb[i]));
		    intree[nghb[i]] = true;
		    nrofnodesintree++;
		}
	}
	if (nrofnodesintree == n/2)
	    return true; // Is a vertex induced tree!
	else
	    return false; // disconnected nodeset
    }

    /**
     * Returns a string representation of the graph in adjacencylist
     * format: <br><pre>
     * 0: nghb1 nghb2 nghb3
     * ... 
     * i: nghbi1 nghbi2 nghbi3
     * ... </pre>
     * @return a String representation of this CubicGraph
     */
    public String toString(){
	String s = n + "\n";
	for(int i = 0; i < adjTab.length; i++)
	    if (!isRemoved(i))
		s+=i+": " + adjTab[i][0] + "\t" + 
		    adjTab[i][1] + "\t" + adjTab[i][2] +"\n";
	return s;
    }

    /**
     * Returns a copy of this graph. Note that this is equivalent to
     * using the Copy Constructor:<br><pre>
     * CubicGraph C = aC.clone();
     * CubicGraph C = new CubicGraph(C);
     * </pre>
     * gives the same result.
     * @return a clone of this CubicGraph
     */
    public Object clone(){
	CubicGraph c = null;
	try{
	    c = (CubicGraph) super.clone();
	    c.adjTab = (int [][]) adjTab.clone();
	    for (int i =0; i < adjTab.length; i++)
		adjTab[i] = (int []) adjTab[i].clone();
	}
	catch(CloneNotSupportedException e){} //should not be possible
	return c;
    }

    /** 
     * Checks if the argument is equal to this object (without
     * checking for ismorphism, just plain equality) following the
     * {@link Graph} interface.
     * @param o - the reference object with which to compare
     * @return true if the specified Object is equal to this
     *         CubicGraph, false otherwise */
    public boolean equals(Object o){
	if (!(o instanceof Graph))
	    return false;
	Graph g = (Graph) o;
	if (nrOfNodes() != g.nrOfNodes() || nrOfEdges() != g.nrOfEdges())
	    return false;
	int i,j;
	NodeIterator ni = first();
	NodeIterator nj = g.first();
	while (ni.hasNext()){
	    i = ni.next();
	    j = nj.next();
	    if (i != j)
		return false;
	    else{
		int [] nghbi = neighbors(i);
		int [] nghbj = g.neighbors(j);
		Arrays.sort(nghbi);
		Arrays.sort(nghbj);
		if (nghbj.length != 3)
		    return false;
		for (int k = 0; k < 3; k++)
		    if (nghbi[k] != nghbj[k])
			return false;
	    }
	}
	return true;
    }
}
