/* CycleCostAlgorithm.java
   ---------
   2002 (c) Dries Van Dyck
   Combinatorial Algorithms Research Group
   Department of Applied Mathematics and Computer Science
   University of Ghent
   Krijgslaan 281--S9
   B-9000 GENT Belgium
*/
import java.util.Vector;
import java.util.ListIterator;
import java.io.PrintStream;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.FileNotFoundException;

/** 
 * This class represents an improved algorithm with additional
 * heuristics on the algorithm using graphical methods by Yutsis,
 * Vanagas et. al for the calculation of general recoupling
 * coefficients of angular momenta.
 * <p>
 * With every edge a cost is associated equal to the difference of the
 * length of the two smallest cycles the edge is part of. The cost of
 * a cycle is defined to be the sum of the cost of the edges in the
 * cycle.
 * <p>
 * This algorithm will reduce the graph in the following way: when
 * there are bubbles present they are removed, otherwise when there are
 * triangles present they are removed. When there are no bubbles nor
 * triangles present the cycle with the smallest cost of all the smallest
 * cycles is made smaller by an interchange on the edge with minimum cost
 * of the cycle.
 * @see Yutsis 
 * @see CycleGenerator
 * @see Cycle
 * @author Dries.VanDyck@rug.ac.be
 */
public class CycleCostAlgorithm{
  /** 
   * The {@link Yutsis} object defining the problem.
   * @see Yutsis
   */
  Yutsis y;
  CycleGenerator cg = null;

  PrintStream log = System.err;  
  int nrofinterchanges = 0;
  /** 
   * Class representing the smallest cycle an {@link Edge} is part of and
   * the cost of the {@link Edge}. 
   */
  class EdgeCost{
    int shortestCycle = y.order();
    int cost = y.order();

    EdgeCost(int cyclelength){
      shortestCycle = cyclelength;
    }

  }
    
  /** 
   * A table with all the {@link Edge} costs (direct acces).
   */
  EdgeCost [] edgeCostTable = null;

  /** 
   * Constructs a new CycleCostAlgorithm object for the {@link
   * Yutsis} y.
   * @param y a {@link Yutsis} object 
   * @see Yutsis
   */
  public CycleCostAlgorithm(Yutsis y){
    this.y = y;
  }

  /** 
   * @return the number of interchanges needed to reduce the Graph.
   */
  public int nrOfInterchanges() { return nrofinterchanges; }
  
  /** 
   * Returns an EdgeCost object for the {@link Edge} <code>(i,j)</code>.
   * @param i endpoint of the {@link Edge}
   * @param j endpoint of the {@link Edge}
   * @return the cost of the {@link Edge} specified by its endpoints
   */
  EdgeCost edgeCost(int i, int j){
    if (i > j)
      return edgeCost(j,i);
    else if (i == j)
      return null;
    return edgeCostTable[j*(j-1)/2+i];
  }

  private void setEdgeCost(int i, int j, EdgeCost ec){
    if (i > j){
      setEdgeCost(j,i,ec);
      return;
    }
    else if (i == j)
      throw new IllegalArgumentException("i == j");
    edgeCostTable[j*(j-1)/2+i] = ec;
  }

  private void clearEdgeCosts(){
    for (int i=0; i < y.order()*(y.order()-1)/2; i++)
      edgeCostTable[i] = null;
  }
  
  /** 
   * Calculates the edge costs for the graph. The cost of an edge
   * is the difference in length of it's two shortest cycles in which
   * it is involved.
   */
  void calculateEdgeCosts(){
    if (edgeCostTable == null)
      edgeCostTable = new EdgeCost[y.order()*(y.order()-1)/2];
    else 
      clearEdgeCosts();
    for (int i=cg.girth(); i < y.order(); i++){
      Vector c_i = cg.cycles(i);
      if (c_i != null){
	ListIterator li = c_i.listIterator();
	while (li.hasNext()){
	  Cycle c = (Cycle) li.next();
	  NodeIterator ni = c.first();
	  int first = ni.next();
	  int previous = first;
	  while (ni.hasNext()){
	    int current = ni.next();
	    updateEdgeCost(previous,current,c);
	    previous = current;
	  }
	  updateEdgeCost(previous,first,c);
	}
      }
    }
  }

  private void updateEdgeCost(int previous, int current, Cycle c){
    EdgeCost ec = edgeCost(previous,current);
    if (ec == null)
      setEdgeCost(previous,current,new EdgeCost(c.length()));
    else if (ec.shortestCycle > c.length()){
      ec.cost = ec.shortestCycle - c.length();
      ec.shortestCycle = c.length();
    }
    else if (ec.cost > c.length() - ec.shortestCycle) // added condition 
      ec.cost =  c.length() - ec.shortestCycle;
  }

  /** 
   * Searches for the best operation to reduce the {@link Yutsis} and
   * performs it.  If there is a bubble or triangle in the graph it
   * will remove the bubble or triangle. Otherwise it will use a {@link
   * CycleGenerator} to generate all relevant cycles.  Afterwards it
   * calculates all edgecosts and searches for the smallest {@link
   * Cycle} with the edge with minimal cost.  When there are multiple
   * cycles of this kind the one with lowest cost is choosen.
   * @see CycleCostAlgorithm#bestCycle 
   */
  public void performOperation(){
    int [] bubble = y.bubble();  
    if (bubble[0] != -1){
      log("Best Cycle: " + bubble[0] + "->" + bubble[1] + "->"+bubble[0]);
      y.removeBubble(bubble);
      return;
    }
    nrofinterchanges++; // triangle counts for an interchange
    int [] triangle = y.triangle();
    if (triangle[0] != -1){
      log("Best cycle: "+ 
	  triangle[0]+"->"+triangle[1]+"->"+triangle[2]+"->"+triangle[0]);
      y.removeTriangle(triangle);
      return;
    }
    int [] bestcycleedge = {-1,-1};
    Cycle bestcycle = bestCycle(bestcycleedge);
    printCycleCosts(log);
    log("Best Cycle:"+bestcycle+"; best edge: "
	+ (y instanceof YutsisGraph ? 
	   ((YutsisGraph) y).edges(bestcycleedge[0],
				   bestcycleedge[1])[0].toString() 
	   : "(" +bestcycleedge[0]+","+bestcycleedge[1]+")"));
    reduceCycle(bestcycle,bestcycleedge);
  }

  /**
   * Prints all considered {@link Cycle}s with the costs of their edges on
   * System.out.
   * @see Cycle
   */  
  public void printCycleCosts(){
    printCycleCosts(System.out);
  }  

  /** 
   * Prints all considered {@link Cycle}s with the costs of their edges on
   * the given PrintStream.
   * @param out the PrintStream to print to
   * @see Cycle
   */
  public void printCycleCosts(PrintStream out){
    if (out == null)
      return;  
    for (int i=cg.girth(); i < y.order(); i++){
      Vector c_i = cg.cycles(i);
      if (c_i != null){
	out.println(c_i.size()+" cycles of length " + i);
	ListIterator li = c_i.listIterator();
	while (li.hasNext()){
	  Cycle c = (Cycle) li.next();
	  printCycleCost(c,out);
	  out.println();
	}
      }
    }
  }
  
  /** 
   * Prints the given {@link Cycle} with the costs of his edges on
   * the given PrintStream.
   * @param c the Cycle to printed
   * @param out the PrintStream to print to
   * @see Cycle
   */
  public void printCycleCost(Cycle c, PrintStream out){
    int previous, current, first;
    NodeIterator ni = c.first(); 
    first = current = ni.next();
    out.print(current);
    while (ni.hasNext()){
      previous = current;
      current = ni.next();
      if (edgeCost(previous,current).cost != y.order())
	  out.print("-"+edgeCost(previous,current).cost+"->"+current);
      else
	  out.print("-!!"+edgeCost(previous,current).cost+"!!->"+current);
    }
    if (edgeCost(current,first).cost != y.order())
	out.print("-"+edgeCost(current,first).cost+"->"+first);
    else
	out.print("-!!"+edgeCost(current,first).cost+"!!->"+first);
  }

  /** 
   * This function returns the best {@link Cycle} to reduce and fills
   * the best edge to interchange out of the cycle (min length 4) It
   * will use a {@link CycleGenerator} to generate all relevant
   * cycles. Afterwards it calculates all edgecosts and searches for
   * the smallest cycle with the edge with minimal cost.  When there
   * are multiple cycles of this kind the one with lowest cost is
   * returned.
   * @param bestcycleedge array where the best edge to be interchanged 
   *                      out of the {@link Cycle} will be filled in
   * @return the best {@link Cycle}
   * @see CycleCostAlgorithm#performOperation 
   */
  public Cycle bestCycle(int [] bestcycleedge){
    if (cg == null)
      cg = new CycleGenerator(y);
    else
      cg.regenerate();
    if (cg.girth() == 3){
	return (Cycle) cg.cycles(3).firstElement();
    }
    calculateEdgeCosts();
    Cycle bestcycle = null;
    int mincyclecost = y.order()*y.order();
    int bestcycleedgecost = y.order();
    int maxnrofbestedges = 0;
    Vector girthCycles = cg.cycles(cg.girth());
    ListIterator li = girthCycles.listIterator();
    while (li.hasNext()){
      Cycle c = (Cycle) li.next();
      NodeIterator ni = c.first();
      int first = ni.next();
      int previous = first;
      int cyclecost = 0;
      int [] bestedge = {-1,-1};
      int minedgecost = y.order();
      int nrofbestedges = 0;
      while (ni.hasNext()){
	int current = ni.next();
	int edgecost = edgeCost(previous,current).cost;
	cyclecost += edgecost;
	if (edgecost < minedgecost){
	  minedgecost = edgecost;
	  bestedge[0] = previous;
	  bestedge[1] = current;
	  nrofbestedges = 1;
	} 
	else if (edgecost == minedgecost)
	    nrofbestedges++;
	previous = current;
      }
      int edgecost = edgeCost(previous,first).cost;
      cyclecost += edgecost;
      if (edgecost < minedgecost){
	minedgecost = edgecost;
	bestedge[0] = previous;
	bestedge[1] = first;
      } else if (edgecost == minedgecost)
	  nrofbestedges++;
      if (minedgecost < bestcycleedgecost ||
	  (minedgecost == bestcycleedgecost 
	   && nrofbestedges < maxnrofbestedges) ||
	  (minedgecost == bestcycleedgecost 
	   && nrofbestedges == maxnrofbestedges && cyclecost < mincyclecost)){
	// take in account nr of times minedgecost is present?? 
	bestcycleedgecost = minedgecost;
	bestcycleedge[0] = bestedge[0];
	bestcycleedge[1] = bestedge[1];
	maxnrofbestedges = nrofbestedges;
	mincyclecost = cyclecost;
	bestcycle = c;
      }      
    }
    return bestcycle;
  }  

  /**
   * This method returns the nodes to be interchanged from the
   * <code>bestcycledge</code> in order to reduce the 
   * <code>bestcycle</code> one unit in length, without applying
   * the interchange itself. 
   * @param bestcycle the {@link Cycle} to be reduced
   * @param bestcycleedge the edge on which to apply the interchange 
   * @return the nodes to interchange of the edge <code>bestcycleedge</code>
   *         in order to reduce the cycle one unit in length.
   * @see Cycle
   */
  public int [] interchangeNodes(Cycle bestcycle, int [] bestcycleedge){
    // Best cycle and his best edge found -> reduce it's size together with
    // his neighbour cycle sharing the edge
    int [] icnodes = {-1,-1};
    NodeIterator ni = bestcycle.nodeIterator(bestcycleedge[0]);
    int [] nghbcycle = {-1,-1};
    if (ni.hasPrevious()){
      nghbcycle[0] = ni.previous();
      ni.next();
    }
    else 
      nghbcycle[0] = bestcycle.last().previous();
    ni.next();
    if (ni.hasNext()){
      ni.next();
      nghbcycle[1] = ni.hasNext() ? ni.next() : bestcycle.first().next();
    }
    else {
      ni = bestcycle.first();
      ni.next();
      nghbcycle[1] = ni.next();
    }
    icnodes[0] = nghbcycle[0]; 
    int [] nghb = y.neighbors(bestcycleedge[1]);
    for (int i = 0; i < 3; i++)
      if (nghb[i] != nghbcycle[1] && nghb[i] != bestcycleedge[0])
	icnodes[1] = nghb[i];
    return icnodes;
  }
    
  private void reduceCycle(Cycle bestcycle, int [] bestcycleedge){
    int [] icnodes = interchangeNodes(bestcycle, bestcycleedge);
    y.interchange(bestcycleedge,icnodes);
  }

  /** 
   * Returns the girth of the {@link Yutsis}.
   * @return the girth of the {@link Yutsis}
   */
  public int girth() { 
      if (cg == null)
	  cg = new CycleGenerator(y);
      else
	  cg.regenerate(); 
      return cg.girth();
  }   

  /** 
   * Returns all girthcycles as a Vector. This is a subset of the relevant
   * {@link Cycle}s generated by {@link CycleGenerator}.
   * @return the girth {@link Cycle}s of the {@link YutsisGraph} as a Vector
   * @see Cycle
   * @see CycleGenerator
   */
  public Vector girthCycles(){
    if (cg == null)
      cg = new CycleGenerator(y);
    else
      cg.regenerate();
    return cg.cycles(cg.girth());
  }

  protected void log(String tolog){
    if (log == null)
      return;
    log.println(tolog);
  }

  /** 
   * Sets the PrintStream to which all operation which alter the
   * graph will be logged. Default logstream is <code>System.err</code>.
   * When no logging is desired, set it to be <code>null</code>.
   * @param logstream The PrintStream where logging will be done
   */
  public void setLogStream(PrintStream logstream){
    log = logstream;
  }


  /** 
   * Reduces the {@link YutsisGraph} to a so called triangular delta
   * by repeatedly calling {@link CycleCostAlgorithm#performOperation}.
   */
  public void reduce(){
    while (!y.triangularDelta()){
      performOperation();
    }
  }  

  /** 
   * This program reduces a Yutsis Graph to a so called triangular
   * delta in as few operations possible, while constructing the
   * general recoupling coefficient associated with it, which is
   * written to the screen.
   */
  public static void main(String [] args) throws IOException{
      String usage = 
	  "Usage: java [-options] CycleCostAlgorithm <YutsisGraph>"
	  +"\n <YutsisGraph>:= \"<Bra|Ket>\" | <filename>\n <filename> must be in BRAKET of YTS format"
	  + "\n options:   v: verbose output (same as \"-or\")"
	  + "\n            o: output graph operations"
	  + "\n            r: output rule selection";
      if (args.length < 1){
	  System.err.println(usage);
	  return;
      }
      boolean outputoperations = false;
      boolean outputrules = false;
      for (int i = 0; i < args.length-1; i++)
	  if (args[i].charAt(0) == '-'){
	      for (int k = 1; k < args[i].length(); k++)
		  switch (args[i].charAt(k)){
		  case 'v':
		      outputoperations = outputrules  = true;
		      break;
		  case 'o':
		      outputoperations = true;
		      break;
		  case 'r': 
		      outputrules = true;
		      break;
		  default:
		      System.err.println("Unknown option -" + 
					 args[i].charAt(k) + "\n" + usage);
		      return;
		  }
	  }
      YutsisGraph y;
      try {	  
	  if (args[args.length-1].indexOf('|') != -1){
	      y = new YutsisGraph(args[args.length-1]);
	  }
	  else {
	      int format = YutsisGraph.guessFileFormat(args[args.length-1]);
	      if (format == -1){
		  System.err.println(usage);
		  return;
	      }
	      y = new YutsisGraph(new BufferedReader(new FileReader(args[args.length-1])),
				  format);
	  }
      }
      catch (FileNotFoundException e){
	  System.err.println("File " + args[args.length-1] + " not found.");
	  return;
      }
      catch (IllegalArgumentException e){
	  System.err.println(e.getMessage());
	  return;
      }
      y.setLogStream(outputoperations ? System.out : null);
      reduceYutsisGraph(y, outputrules ? System.out : null);
  }

  /** 
   * This function calls {@link CycleCostAlgorithm#performOperation}
   * until the {@link YutsisGraph} is equal to a triangular delta and
   * prints the formula.
   */
  public static void reduceYutsisGraph(Yutsis y, PrintStream ruleslogstream){
    CycleCostAlgorithm cca = new CycleCostAlgorithm(y);
    cca.setLogStream(ruleslogstream);
    cca.reduce();
    if (y instanceof YutsisGraph){
	YutsisGraph yg = (YutsisGraph) y;
	System.out.println("#summations: " + 
			   yg.genRecoupCoeff().nrOfSummations()
			   + " #6j's: " + yg.genRecoupCoeff().nrOf6js());
	System.out.println(yg.genRecoupCoeff());
    }
    else
	System.out.println("#ic's cca:"+cca.nrOfInterchanges());
  }
}
