/* GRWrappedStringVisitor.java
   ---------
   2001 (c) Dries Van Dyck
   Combinatorial Algorithms Research Group
   Department of Applied Mathematics and Computer Science
   University of Ghent
   Krijgslaan 281--S9
   B-9000 GENT Belgium
*/

import java.util.Enumeration;
import java.awt.FontMetrics;

/** 
 * A GRVisitor which constructs a wrapped String representation
 * of an object from the GR-family, including {@link GenRecoupCoeff}.
 * When a FontMetrics Class is specified in the constructor, the
 * wrapsize is in pixels and takes in account the used font.
 * @see GRVisitor 
 * @see GenRecoupCoeff 
 * @see GRCompoundFactor 
 * @see GRSummation 
 * @see GRFactor 
 * @see GRPreFactor 
 * @see GRKrDelta 
 * @see GR6jSymbol
 * @author Dries.VanDyck@rug.ac.be 
 */
public class GRWrappedStringVisitor extends GRVisitor{
    
    private StringBuffer wrappedstring = new StringBuffer();
    private FontMetrics fm = null;
    private int wrapsize = 80;
    private int wrapcount = 0;
    
    /** 
     * Constructs a new GRWrappedStringVisitor. Default wrapsize is 80
     * characters.
     */
    public GRWrappedStringVisitor(){}

    /** 
     * Constructs a new GRWrappedStringVisitor, with given wrapsize 
     * in pixels and the FontMetrics class for the used font.
     * @param wrapsize number of allowed pixels on a line
     * @param FontMetrics the FontMetrics for the used font
     */
    public GRWrappedStringVisitor(int wrapsize, FontMetrics fm){
	this.wrapsize = wrapsize;
	this.fm = fm;
    }

    /** 
     * Constructs a new GRWrappedStringVisitor which wraps at
     * <code>wrapsize</code> characters.
     * @param wrapsize number of allowed chars on one line
     */
    public GRWrappedStringVisitor(int wrapsize){ this.wrapsize = wrapsize; }

    /** 
     * Specific method for visiting a {@link GenRecoupCoeff} object.
     * @param grc the {@link GenRecoupCoeff} to be visited 
     * @see GenRecoupCoeff
     */
    public void visitGenRecoupCoeff(GenRecoupCoeff grc){}

    /** 
     * Specific method for visiting a {@link GRCompoundFactor} object.
     * @param grcf the {@link GRCompoundFactor} to be visited 
     */
    public void visitGRCompoundFactor(GRCompoundFactor grcf){}

    /** 
     * Specific method for visiting a {@link GRSummation} object.
     * @param grs the {@link GRSummation} to be visited 
     * @see GRSummation
     */
    public void visitGRSummation(GRSummation grs){
	String [] couplings = grs.couplings();
	String summation = "sum("+grs.summVar()+
	    ",max(|"+couplings[0]+"-"+couplings[1]+"|,|"+
	    couplings[2]+"-"+couplings[3]+"|)..min("+
	    couplings[0]+"+"+couplings[1]+","+
	    couplings[2]+"+"+couplings[3]+"))";
	wrap();
	append(summation);
    }

    /** 
     * Specific method for visiting a {@link GRPreFactor} object.
     * @param grpf the {@link GRPreFactor}to be visited 
     * @see GRPreFactor
     */
    public void visitGRPreFactor(GRPreFactor grpf){
	String prefactor = grpf.toString();
	if (wrapcount+size(prefactor) >= wrapsize){
	    if (size(prefactor) < wrapsize){
		wrap();
		append(prefactor);
	    }
	    else if (prefactor.startsWith("(-1)^(")){
		int productindex = prefactor.indexOf(')',5)+1;
		String exp = prefactor.substring(0,productindex);
		String prod = prefactor.substring(productindex);
		if (wrapcount+size(exp) < wrapsize)
		    append(exp);
		else
		    wrapExp(exp);
		if (wrapcount+size(prod) < wrapsize)
		    append(prod);
		else
		    wrapProduct(prod);
	    }
	    else 
		wrapProduct(prefactor);
	}
	else 
	    append(prefactor);
    }

    /** 
     * Specific method for visiting a {@link GRKrDelta} object.
     * @param grcd the {@link GRKrDelta} to be visited 
     * @see GRKrDelta
     */
    public void visitGRKrDelta(GRKrDelta grcd){
	String delta = grcd.toString();
	if (wrapcount+size(delta) >= wrapsize)
	    wrap();
	append(delta);

    }

    /** 
     * Specific method for visiting a {@link GR6jSymbol} object.
     * @param grpf the {@link GR6jSymbol} to be visited 
     * @see GR6jSymbol
     */
    public void visitGR6jSymbol(GR6jSymbol grpf){
	String w6j = grpf.toString();
	if (wrapcount+size(w6j) >= wrapsize)
	    wrap();
	append(w6j);
    }

    /** 
     * Returns the wrapped String representing the object visited.
     * @return the resulting wrapped String
     */
    public String result(){
	return wrappedstring.toString();
    }

    private void wrap(){
	wrappedstring.append('\n');
	wrapcount = 0;
    }

    private void append(String s){
	wrappedstring.append(s);
	wrapcount += size(s);
    }

    private void wrapProduct(String product){
	if (fm == null) {
	    int pos = wrapsize - wrapcount;
	    int lastwrap = 0;
	    while (pos < product.length()){
		pos = product.lastIndexOf("*(",pos);
		append(product.substring(lastwrap,pos));
		lastwrap = pos;
		wrap();
		pos += wrapsize;
	    }
	    append(product.substring(lastwrap));
	}
	else {
	    int pos = 0;
	    int lastwrap = 0;
	    int wrappos = 0;
	    while (pos < product.length()){
		do {
		    wrappos = pos;
		    pos = product.indexOf("*(",pos+1);
		    pos = pos != -1 ? pos : product.length();
		} while (pos < product.length() && wrapcount + 
			 size(product.substring(lastwrap,pos)) 
			 < wrapsize);
		append(product.substring(lastwrap,wrappos));
		if (pos != product.length()){
		    lastwrap = wrappos;
		    wrap();
		}
		else if (wrappos < pos){
		    if (wrapcount + size(product.substring(wrappos,pos))
			> wrapsize)
			wrap();
		    append(product.substring(wrappos,pos));
		}
	    }
	}
    }

    private void wrapExp(String exp){
	if (fm == null){
	    int pos = wrapsize-wrapcount;
	    int lastwrap = 0;
	    while (pos < exp.length()){
		pos = Math.max(exp.lastIndexOf('+',pos),
			       exp.lastIndexOf('-',pos));
		append(exp.substring(lastwrap,pos));
		lastwrap = pos;
		wrap();
		pos+=wrapsize;
	    }
	    append(exp.substring(lastwrap));
	}
	else {
	    int posplus = 0;
	    int posmin = 0;
	    int pos = 0;
	    int lastwrap = 0;
	    int wrappos = 0;
	    
	    while (pos < exp.length()){
		do {
		    wrappos = pos;
		    posplus = exp.indexOf('+',pos+1);
		    posplus = posplus != -1 ? posplus : exp.length();
		    posmin = exp.indexOf('-',pos+1);
		    posmin = posmin != -1 ? posmin : exp.length();
		    pos = Math.min(posmin,posplus);
		} while (pos < exp.length() 
			 && wrapcount+size(exp.substring(lastwrap,pos))
			 < wrapsize);
		append(exp.substring(lastwrap,wrappos));
		if (pos != exp.length()){
		    lastwrap = wrappos;
		    wrap();
		}
		else if (wrappos < pos){
		    if (wrapcount + size(exp.substring(wrappos,pos))
			> wrapsize)
			wrap();
		    append(exp.substring(wrappos,pos));
		}
	    }
	}
    }

    private int size(String s){
	return fm == null ? s.length() : fm.stringWidth(s);
    }
}
