/* GYutsisApplet.java
   2002 (c) Dries Van Dyck
   Combinatorial Algorithms Research Group
   Department of Applied Mathematics and Computer Science
   University of Ghent
   Krijgslaan 281--S9
   B-9000 GENT Belgium
*/
import java.util.NoSuchElementException;
import java.util.Vector;

import java.io.OutputStream;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.FileNotFoundException;
import java.io.IOException;

import javax.swing.*;
import javax.swing.event.*;

import java.awt.*;
import java.awt.event.*;

/** 
 * This class implements an Applet version of the GYutsis program
 * illustrating the algorithm implemented by {@link
 * CycleCostAlgorithm}. In advanced mode the algorithm can be
 * followed step by step while keeping track of the underlying Yutsis
 * Graph.
 * @see GYutsis
 * @see YutsisGraph 
 * @see GenRecoupCoeff 
 * @see StreamPanel
 * @author Dries.VanDyck@rug.ac.be 
 */
public class GYutsisApplet extends JApplet 
    implements ActionListener, ChangeListener, ComponentListener{
    static final String VERSION = "1.0";
    static final int OUTPUTROWS = 8;
    static final int OUTPUTCOLUMNS = 50;
    static final String ABOUTTEXT = 
	"GYutsisApplet 1.0 (Applet version of GYutsis 1.0)\n"+ 
	" Copyright (c) 2002 Dries Van Dyck\n"+
	" Author: Dries.VanDyck@rug.ac.be\n" + 
	"             Department of Applied Mathematics & Computer Science\n"+
	"             University of Ghent\n"+
	"             Krijgslaan 281 -- S9, Ghent, Belgium\n"+
	" License: free for academic use, otherwise contact author";
    static final String EXAMPLE_F8 =
	"<(((j1,j2)j9,j3)j10,(((j4,j5)j11,j6)j12,(j7,j8)j13)j14)j15|(((j1,j4)j16,j7)j17,(((j2,j5)j18,(j8,j3)j19)j20,j6)j21)j15>";
    
    // Underlying model
    YutsisModel y;
    CycleCostAlgorithm cca;
    boolean reduced = false;
    // IO panel
    JPanel iopanel;
    JTextField input;
    JButton reducebutton;
    JScrollPane scrollpane;
    JTextArea output;
    JTextField nrofsummations;
    JTextField nrof6js;
    //Advanced Panel
    boolean advancedmode = false;
    JPanel advancedpanel;
    StreamPanel rules, operations, user;
    PrintStream userstream;
    //Advanced Buttons
    JButton graph, cycles, bestcycle, step;
    //Menus
    JMenuBar menubar;
    JMenu operationsmenu, helpmenu;
    //Operations Menu
    JMenuItem graphmi, cyclesmi, bestcyclemi, stepmi, reducemi;
    //Help Menu
    JMenuItem aboutmi;
    
    /** 
     * Creates the contentpane of this applet.
     */
    public void init(){
	advancedmode = "true".equals(getParameter("advancedmode"));
	if (advancedmode){
	    JPanel iopanel = new JPanel(new GridBagLayout());
	    createIOPanel(iopanel);
	    createAdvancedPanel();
	    setContentPane(new JSplitPane(JSplitPane.VERTICAL_SPLIT,
					  iopanel,advancedpanel));
	}
	else{
	    getContentPane().setLayout(new GridBagLayout());
	    createIOPanel(getContentPane());
	}
	createMenus();
    }

    private void createIOPanel(Container panel){
	// Yutsis Graph: label
	GridBagConstraints gbc = new GridBagConstraints();
	gbc.insets = new Insets(4,2,2,4);
	gbc.gridx=0; gbc.gridy=0;
	gbc.gridwidth = 1; gbc.gridheight= 1;
	panel.add(new JLabel("Braket:"),gbc);
	// Input textfield
	input = new JTextField(40);
	input.setToolTipText("Fill in BRAKET of a Yutsis Graph");
	input.setText(EXAMPLE_F8);
	input.selectAll();
	input.addActionListener(this);
	gbc.gridwidth = 2; gbc.gridx = 1;
	gbc.weightx = 1; 
	gbc.fill = GridBagConstraints.HORIZONTAL;
	panel.add(input,gbc);
	// Reduce button
	reducebutton = new JButton("Reduce");
	reducebutton.setEnabled(false);
	reducebutton.setToolTipText("Reduce the Yutsis Graph to a triangular delta");
	reducebutton.addActionListener(this);
	gbc.gridx = 3; gbc.gridwidth = 1;
	gbc.weightx = 0; 
	gbc.fill = GridBagConstraints.NONE;
	panel.add(reducebutton,gbc);
	// Summation Formula: label
	gbc.gridx = 0; gbc.gridy = 1;
	gbc.gridwidth = 2; gbc.anchor = GridBagConstraints.NORTH;
	gbc.weighty = 0.5;
	panel.add(new JLabel("Summation Formula:"),gbc);
	// #Summations: label 
	gbc.anchor = GridBagConstraints.SOUTH; 
	gbc.gridy = 2; gbc.gridwidth = 1;
	panel.add(new JLabel("#Summations:"),gbc);
	// #Summations: textfield
	nrofsummations = new JTextField(3);
	nrofsummations.setEditable(false);
	gbc.fill = GridBagConstraints.HORIZONTAL;
	gbc.anchor = GridBagConstraints.SOUTHWEST;
	gbc.gridx = 1; gbc.gridy = 2;
	panel.add(nrofsummations,gbc);
	// #Wigner 6j's: label
	gbc.fill = GridBagConstraints.NONE;
	gbc.anchor = GridBagConstraints.NORTH;
	gbc.gridx = 0; gbc.gridy = 3; 
	gbc.weighty = 0.5;
	gbc.gridwidth = 1;
	panel.add(new JLabel("#Wigner 6j's:"),gbc);
	// #Wigner 6j's: textfield
	nrof6js = new JTextField(3);
	nrof6js.setEditable(false);
	gbc.fill = GridBagConstraints.HORIZONTAL;
	gbc.anchor = GridBagConstraints.NORTHWEST;
	gbc.gridx = 1; gbc.gridy = 3;
	panel.add(nrof6js,gbc);
	// Output textarea
	output = new JTextArea(OUTPUTROWS,OUTPUTCOLUMNS);
	output.setEditable(false);
	gbc.insets = new Insets(0,0,0,0);
	gbc.gridx = 2; gbc.gridy = 1;
	gbc.gridwidth = 2; gbc.gridheight = 3;
	gbc.fill = GridBagConstraints.BOTH;
	gbc.weightx = 1; gbc.weighty = 1;
	scrollpane = new JScrollPane(output,
				     JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
				     JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
	scrollpane.addComponentListener(this);
	panel.add(scrollpane,gbc);
    }

    private void createAdvancedPanel(){
	advancedpanel = new JPanel(new BorderLayout());
	// Create TabbedPane
	JTabbedPane tabpane = new JTabbedPane();
	user = new StreamPanel(15,30,"User commands output:",false);
	userstream = user.getPrintStream();
	tabpane.add("User",user);
	rules = new StreamPanel(15,30,"Rule selection output:",false);
	tabpane.add("Rules",rules);
	operations = new StreamPanel(15,30,"Graph operations output:",false);
	tabpane.add("Operations",operations);
	tabpane.setSelectedIndex(0);
	advancedpanel.add(tabpane,BorderLayout.CENTER);
	// Create Panel with user operation buttons
	GridBagLayout gbl = new GridBagLayout();
	JPanel buttonpanel = new JPanel(gbl);
	// Set GridBag Constraints
	GridBagConstraints gbc = new GridBagConstraints();
	gbc.gridx = 0;
	gbc.gridy = 0;
	gbc.gridheight = 1;
	gbc.gridwidth = 1;
	gbc.fill = GridBagConstraints.HORIZONTAL;
	gbc.anchor = GridBagConstraints.CENTER;
	JLabel userlabel = new JLabel("User operations:");
	gbl.setConstraints(userlabel,gbc);
	buttonpanel.add(userlabel,gbc);
	// add buttons
	graph = new JButton("Print Graph");
	graph.setEnabled(false);
	gbc.insets = new Insets(16,0,0,0); //top, left, bottom, right
	gbc.gridy = gbc.gridy+1;
	graph.addActionListener(this);
	gbl.setConstraints(graph,gbc);
	buttonpanel.add(graph,gbc);
	cycles = new JButton("Cycles");
	cycles.setToolTipText("Show all girth cycles");
	cycles.setEnabled(false);
	cycles.addActionListener(this);
	gbc.gridy = gbc.gridy+1;
	gbc.gridheight = 1;
	gbl.setConstraints(cycles,gbc);
	buttonpanel.add(cycles,gbc);
	//JButton formula;
	bestcycle = new JButton("Best cycle");
	bestcycle.setToolTipText("Show the best cycle and the best edge to interchange (not for triangles)");
	bestcycle.setEnabled(false);
	bestcycle.addActionListener(this);
	gbc.insets = new Insets(0,0,0,0); //top, left, bottom, right
	gbc.gridy = gbc.gridy+1;
	gbl.setConstraints(bestcycle,gbc);
	buttonpanel.add(bestcycle,gbc);
	step = new JButton("Step");
	step.setToolTipText("Perform one step in the reduction algorithm");
	step.setEnabled(false);
	step.addActionListener(this);
	gbc.gridy = gbc.gridy+1;
	gbl.setConstraints(step,gbc);
	buttonpanel.add(step,gbc);
	advancedpanel.add(buttonpanel,BorderLayout.EAST);
    }
    
    private void createMenus(){
	menubar = new JMenuBar();
	setJMenuBar(menubar);
	//Operations menu
	operationsmenu = new JMenu("Operations");
	reducemi = new JMenuItem("Reduce",'R');
	reducemi.addActionListener(this);
	reducemi.setEnabled(false);
	menubar.add(operationsmenu);
	if (advancedmode){
	    //Additional operations menu item when in advanced mode
	    graphmi = new JMenuItem("print Graph", 'G');
	    graphmi.addActionListener(this);
	    graphmi.setEnabled(false);
	    cyclesmi = new JMenuItem("girth Cycles", 'C');
	    cyclesmi.addActionListener(this);
	    cyclesmi.setEnabled(false);
	    bestcyclemi = new JMenuItem("Best cycle",'B');
	    bestcyclemi.addActionListener(this);
	    bestcyclemi.setEnabled(false);
	    stepmi = new JMenuItem("Step", 'S');
	    stepmi.addActionListener(this);
	    stepmi.setEnabled(false);
	    operationsmenu.add(graphmi);
	    operationsmenu.addSeparator();
	    operationsmenu.add(cyclesmi);
	    operationsmenu.add(bestcyclemi);
	    operationsmenu.addSeparator();
	    operationsmenu.add(stepmi);
	}
	operationsmenu.add(reducemi);
	// Help menu on the right
	helpmenu = new JMenu("Help");
	aboutmi = new JMenuItem("About",'A');
	aboutmi.addActionListener(this);
	helpmenu.add(aboutmi);
	menubar.add(Box.createHorizontalGlue());
	menubar.add(helpmenu);
    }

    private void setLogStreams(){
	y.setLogStream(operations.getPrintStream());
	cca.setLogStream(rules.getPrintStream());
    }

    /** 
     * Changes the {@link YutsisModel} y.
     * @param y the new YutsisModel
     * @see YutsisModel
     */
    void setYutsis(YutsisModel y){
	this.y = y;
	cca = new CycleCostAlgorithm(y);
	y.addChangeListener(this);
	boolean notreduced = !y.triangularDelta();
	reducebutton.setEnabled(notreduced);
	reducemi.setEnabled(notreduced);
	if (advancedmode){
	    setLogStreams();
	    user.clear();
	    rules.clear();
	    operations.clear();
	    graph.setEnabled(true);
	    graphmi.setEnabled(true);
	}
	else{
	    y.setLogStream(null);
	    cca.setLogStream(null);
	    output.setText("");
	    nrofsummations.setText("");
	    nrof6js.setText("");
	}
	stateChanged(new ChangeEvent(y));
    }


    private void reduced(){
	reducebutton.setEnabled(false);
	reducemi.setEnabled(false);
	if (advancedmode){
	    cycles.setEnabled(false);
	    cyclesmi.setEnabled(false);
	    bestcycle.setEnabled(false);
	    bestcyclemi.setEnabled(false);
	    step.setEnabled(false);
	    stepmi.setEnabled(false);
	}
    }

    /** 
     * Default implementation of the ComponentListener.
     */
    public void componentHidden(ComponentEvent e){}

    /** 
     * Default implementation of the ComponentListener.
     */
    public void componentMoved(ComponentEvent e){}

    /** 
     * Default implementation of the ComponentListener.
     */
    public void componentShown(ComponentEvent e){}
    
    /** 
     * Implementation of the ComponentListener, reuse 
     * GRWrappedStringVisitor to wrap the formula.
     * @see GRWrappedStringVisitor
     */
    public void componentResized(ComponentEvent e){
	//Listens to the scrollpane...
	if (y != null && (advancedmode || reduced))
	    outputFormula();
    }

    /** 
     * Implementation of the ActionListener interface.
     */
    public void actionPerformed(ActionEvent e){
	if (e.getSource() == reducebutton || e.getSource() == reducemi)
	    reduce();
	else if (e.getSource() == input){
	    String text = input.getText().trim();
	    try{ 
		setYutsis(new YutsisModel(text));
	    }
	    catch (IllegalArgumentException iae){
		JOptionPane.showMessageDialog(this, iae.getMessage(),
					      "Illegal input",
					      JOptionPane.ERROR_MESSAGE);
	    }
	}
	else if (e.getSource() == aboutmi)
	    JOptionPane.showMessageDialog(this, ABOUTTEXT, "About GYutsis 1.0",
					  JOptionPane.INFORMATION_MESSAGE);
	else if(advancedmode){
	    if (e.getSource() == graph || e.getSource() == graphmi)
		printGraph();
	    else if (e.getSource() == cycles 
		     || e.getSource() == cyclesmi)
		printCycles();
	    else if (e.getSource() == bestcycle
		     ||e.getSource() == bestcyclemi)
		printBestCycle();
	    else if (e.getSource() == step || e.getSource() == stepmi)
		step();
	}	
    }

    private void reduce() { cca.reduce();}

    private void printGraph(){ userstream.println(y);}

    private void printCycles(){
	int [] abubble = y.bubble();
	if (abubble[0] != -1)
	    userstream.println("Bubble between nodes " 
			       + abubble[0] + " " + abubble[1] + '\n');
	else {
	    Vector girthcycles = cca.girthCycles();
	    for (int i = 0; i < girthcycles.size(); i++)
		userstream.println(girthcycles.elementAt(i));
	    userstream.println();
	}
    }

    private void printBestCycle(){
	int [] abubble = y.bubble();
	if (abubble[0] != -1)
	    userstream.println("Bubble between nodes " 
			       + abubble[0] + " " + abubble[1]+ '\n');
	else {
	    int [] bestedge = {-1,-1};
	    Cycle c = cca.bestCycle(bestedge);
	    userstream.println("Best cycle: " + c 
			       + (bestedge[0] != -1 ? " ; best edge: " 
				  +  bestedge[0] + " " + bestedge[1] 
				  : "" )+'\n');
	}
    }

    private void step() { cca.performOperation(); }
    
    /** 
     * Implementation of the ChangeListener interface.
     */
    public void stateChanged(ChangeEvent ce){
	reduced = y.triangularDelta();
	if (advancedmode || reduced)
	    outputFormula();
	if (reduced)
	    reduced();
	else if (advancedmode){
	    cycles.setEnabled(true);
	    cyclesmi.setEnabled(true);
	    bestcycle.setEnabled(true);
	    bestcyclemi.setEnabled(true);
	    step.setEnabled(true);
	    stepmi.setEnabled(true);
	}
    }
    
    void outputFormula(){
	GRWrappedStringVisitor v = new GRWrappedStringVisitor
	    (Math.min((int) output.getSize().getWidth(), 
		      (int) scrollpane.getSize().getWidth()),
	     output.getFontMetrics(output.getFont()));
	y.genRecoupCoeff().accept(v);
	output.setText(v.result());
	nrofsummations.setText(""+y.genRecoupCoeff().nrOfSummations());
	nrof6js.setText(""+y.genRecoupCoeff().nrOf6js());
    }
}
