/* YutsisGraph.java
   ---------
   2001 (c) Dries Van Dyck
   Combinatorial Algorithms Research Group
   Department of Applied Mathematics and Computer Science
   University of Ghent
   Krijgslaan 281--S9
   B-9000 GENT Belgium
*/
import java.util.Vector;

import java.io.IOException;
import java.io.BufferedReader;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * This class represents a model wrapped round a YutsisGraph object. When
 * the YutsisGraph is structurally changed a ChangeEvent is fired to 
 * all its registered ChangeListeners.
 * @see YutsisGraph
 * @author Dries.VanDyck@rug.ac.be 
 */
public class YutsisModel extends YutsisGraph{        
    
    private Vector changelisteners = new Vector();

    /** 
     * Constructor creating a YutsisModel from a BufferedReader in 
     * YTS format.
     * @param input the source from where the YutsisModel is read
     */
    public YutsisModel(BufferedReader input) throws IOException{
	super(input,YTS);
    }

    /** 
     * Contructor creating a YutsisModel from a BufferedReader in
     * given format.
     * @param input the source from where the YutsisModel is read
     * @param format the format in which the graph is read
     */
    public YutsisModel(BufferedReader input, int format) throws IOException{
	super(input,format);
    }

    /** 
     * Construcs a new YutsisModel object from the braket in the String.
     * @param braket a String describing the YutsisModel in BRAKET format
     * @see Yutsis#BRAKET
     */
    public YutsisModel(String braket){
	super(braket);
    }

    /** 
     * Adds a ChangeListener to the YutsisGraph. Only structural changes are
     * notified, when a node or edge is inverted no listeners are notified. 
     * @param l the ChangeListener to be added
     */
    public void addChangeListener(ChangeListener l){ changelisteners.add(l); }

    /** 
     * Removes the ChangeListener l from the YutsisGraph.
     * @param l the ChangeListener to be removed
     */
    public void removeChangeListener(ChangeListener l){
	changelisteners.remove(l);
    }
    
    /** 
     * Send a ChangeEvent, whose source is this YutsisGraph, 
     * to each listener.
     */
    protected void fireStateChanged(){
	ChangeEvent ce = new ChangeEvent(this);
	for (int i = 0; i < changelisteners.size(); i++)
	    ((ChangeListener) changelisteners.elementAt(i)).stateChanged(ce);
    }

    /** 
     * Formats and removes the bubble as specified by the {@link
     * Yutsis} interface.
     * @param bubble the bubble to be removed.
     * @see Yutsis
     */
    public void removeBubble(int [] bubble){
	super.removeBubble(bubble);
	fireStateChanged();
    }

    /** 
     * Formats and removes the triangle <code>node1, node2,
     * node3</code>, as specified by the {@link Yutsis} interface.
     * @param node1 first node of the triangle to be removed
     * @param node2 second node of the triangle to be removed
     * @param node3 third node of the triangle to be removed 
     * @see Yutsis
     */
    public void removeTriangle(int [] triangle){
	super.removeTriangle(triangle);
	fireStateChanged();
    }	
    
    /** 
     * Interchange on the edge <code>(node1, node2)</code> interchanging 
     * the edges <code>(node1, icnode1)</code> and 
     * <code>(node2, icnode2)</code>, as specified by the {@link Yutsis}
     * interface.
     * @param node1 first node of the {@link Edge} on which the interchange 
     *              is performed
     * @param node2 second node of the {@link Edge} on which the interchange 
     *              is performed
     * @param icnode1 neighbor of node1 to be interchanged with icnode2
     * @param icnode2 neighbor of node2 to be interchanged with icnode1
     * @see Yutsis
     */
    public void interchange(int [] nodes, int [] icnodes){
	super.interchange(nodes,icnodes);
	fireStateChanged();
    }
}
